# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
""" 
This module implements a simple algorithm to display a graph(mainly a tree)
level by level, using only Tkinter.
"""



# TODO: all LevelTree code is DEPRECATED...remove it after cleanning
# and include all code from graph_layout


class LevelTree(object):
    """ Class to render the Graph in a Canvas. """
    def __init__(self, graph):
        self.DY = 65
        self.DX = 15
        self.FONT = "sans-serif"
        self.FONTSIZE = 9
        self.graph = graph
        self.canvas = None
        
    def setCanvas(self, canvas):
        self.canvas = canvas
        
    def paint(self, createNode=None, createEdge=None, maxLevel=9999, usePositions=False):
        """ Paint the Graph, nodes will be positioned by levels.
        Params:
            canvas: the canvas object to paint the graph.
            createNode: function to build the Item that represents a Node.
                    the Item created should have the following methods:
                        getDimensions: return the width and height
                        moveTo: change the position of the Item
            createEdge: function to build an Edge connection two Nodes
            usePositions: if this is True, use the nodes positions without
                recomputing them.
            
            If createNode and createEdge are None, the default ones will be used,
            that requires the setCanvas method had to be called first.
        """
        self.createNode = createNode or self._defaultCreateNode
        self.createEdge = createEdge or self._defaultCreateEdge
        self.maxLevel = maxLevel
        rootNode = self.graph.getRoot()
        
        if usePositions:
            self._paintNodeWithPosition(rootNode)
            self._paintEdges(rootNode)
        else:            
            self._setLevel(rootNode, 0, None)
            self._paintNodeWithChilds(rootNode, 1)
            m = 9999
            for left, right in rootNode.hLimits:
                m = min(m, left)
            self._createEdges(rootNode, -m + self.DY)
        
    def _setLevel(self, node, level, parent):
        """ Set the level of the nodes. """
        node.level = level
        node.parent = parent
        nextLevel = level + 1
        if nextLevel > self.maxLevel:
            return
        for child in node.getChilds():
            if nextLevel > getattr(child, 'level', 0):
                self._setLevel(child, nextLevel, node)
                 
    def _paintNodeWithChilds(self, node, level):
        y = level * self.DY
        
        self._paintNode(node, y)
        
        if level > self.maxLevel:
            return
        
        childs = [c for c in node.getChilds() if c.parent is node]
        n = len(childs)

        if n > 0:
            # width = (xmax - xmin) / n
            for c in childs:
                self._paintNodeWithChilds(c, level + 1)
                
            if n > 1:
                offset = 0
                for i in range(n-1):
                    sep = self._getChildsSeparation(childs[i], childs[i+1])
                    offset += sep
                    c = childs[i+1]
                    c.offset = offset
                
                total = childs[0].half + offset + childs[-1].half
                half = total/2
                for c in childs:
                    c.offset -= half - childs[0].half
                
            else:
                childs[0].offset = 0
            self._getHLimits(node)
        
    def _defaultCreateNode(self, canvas, node, y):
        """ If not createNode is specified, this one will be used
        by default. 
        """
        if canvas is None:
            raise Exception("method setCanvas should be called before using _defaultCreateNode")
        nodeText = node.getLabel()
        textColor = 'black'
        if nodeText.startswith('Project'):
            textColor = 'white'
        
        return canvas.createTextbox(nodeText, 100, y, bgColor='light blue', textColor=textColor, margin=0)
        
    def _defaultCreateEdge(self, srcItem, dstItem):
        if self.canvas is None:
            raise Exception("method setCanvas should be called before using _defaultCreateEdge")
        self.canvas.createEdge(srcItem, dstItem)
        
    def _paintNode(self, node, y):
        """ Paint a node of the graph.
        Params:
           canvas: the canvas in which to paint.
           node: the node of the graph to be painted.
           y: level in the tree where to paint.
        Returns:
           the create item in the canvas.
        """
        item = self.createNode(self.canvas, node, y)
        node.width, node.height = item.getDimensions()
        node.half = node.width / 2
        node.hLimits = [[-node.half, node.half]]
        node.y = item.y
        node.offset = 0
        # Create link from both sides to reach
        node.item = item 
        item.node = node
        
        return item
    
    def _printHLimits(self, node, msg):
        print("\n=====%s========" % msg)
        print(" dd: %s" % node.t.text.replace('\n', '_'))
        print("  offset: %d, width: %d" % (node.offset, node.width))
        print("  hlimits:")
        for l, r in node.hLimits:
            print("   [%d, %d]" % (l, r))
            
    def _getHLimits(self, node):
        """
        This function will traverse the tree
        from node to build the left and right profiles(hLimits)
        for each level of the tree
        """
        node.hLimits = [[-node.half, node.half]]
        childs = [c for c in node.getChilds() if c.parent is node]
        for child in childs:
            count = 1
            if not hasattr(child, 'hLimits'):
                print("node %s has no hLimits" % child.label)
                raise Exception()
            
            for l, r in child.hLimits:
                l += child.offset
                r += child.offset
                if count < len(node.hLimits):
                    if l < node.hLimits[count][0]:
                        node.hLimits[count][0] = l
                    if r > node.hLimits[count][1]:
                        node.hLimits[count][1] = r
                else:
                    node.hLimits.append([l, r])
                count += 1
                
    def _getChildsSeparation(self, child1, child2):
        """ Calcualte separation between siblings
        at each height level. """
        sep = 0 
        hL1 = child1.hLimits
        hL2 = child2.hLimits
        n1 = len(hL1)
        n2 = len(hL2)
        h = min(n1, n2)
            
        for i in range(h):
            right = hL1[i][1]
            left = hL2[i][0]            
            if left + sep < right:
                sep = right - left                
  
        return sep + self.DX
    
    def _createEdges(self, node, x):
        """ Adjust the position of the nodes
        and create the edges between them.
        """
        nx = x + node.offset
        node.item.moveTo(nx, node.y)
        
        if node.level == self.maxLevel:
            return 
        
        for c in node.getChilds():
            if c.parent is node:
                self._createEdges(c, nx)
            self.createEdge(node.item, c.item)
            
    def _paintNodeWithPosition(self, node):
        """ Paint nodes using its position. """
        self._paintNode(node, None)
        
        for child in node.getChilds():
            # parent = None for nodes that have been not traversed
            parent = getattr(child, 'parent', None)
            if parent is None:
                child.parent = node
                self._paintNodeWithPosition(child)
                
    def _paintEdges(self, node):
        """ Paint only the edges between nodes, asumming they are 
        already well positioned. 
        """
        for child in node.getChilds():
            if child.parent is node:
                self._paintEdges(child)        
            self.createEdge(node.item, child.item)






# Copyright (c) 2002-2005 ActiveState Corp.
# See LICENSE.txt for license details.
# Author:
#   Trent Mick (TrentM@ActiveState.com)
# Home:
#   http://trentm.com/projects/which/

"""Find the full path to commands.

which(command, path=None, verbose=0, exts=None)
    Return the full path to the first match of the given command on the
    path.

whichall(command, path=None, verbose=0, exts=None)
    Return a list of full paths to all matches of the given command on
    the path.

whichgen(command, path=None, verbose=0, exts=None)
    Return a generator which will yield full paths to all matches of the
    given command on the path.
    
"""

import os
import sys
import stat
# ---- exceptions


class WhichError(Exception):
    pass

# ---- internal support stuff

def _getRegisteredExecutable(exeName):
    """Windows allow application paths to be registered in the registry."""
    registered = None
    if sys.platform.startswith('win'):
        if os.path.splitext(exeName)[1].lower() != '.exe':
            exeName += '.exe'
        import winreg
        try:
            key = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\" +\
                  exeName
            value = winreg.QueryValue(winreg.HKEY_LOCAL_MACHINE, key)
            registered = (value, "from HKLM\\"+key)
        except winreg.error:
            pass
        if registered and not os.path.exists(registered[0]):
            registered = None
    return registered


def _samefile(fname1, fname2):
    if sys.platform.startswith('win'):
        return (os.path.normpath(os.path.normcase(fname1)) ==
                os.path.normpath(os.path.normcase(fname2)))
    else:
        return os.path.samefile(fname1, fname2)


def _cull(potential, matches, verbose=0):
    """Cull inappropriate matches. Possible reasons:
        - a duplicate of a previous match
        - not a disk file
        - not executable (non-Windows)
    If 'potential' is approved it is returned and added to 'matches'.
    Otherwise, None is returned.
    """
    for match in matches:  # don't yield duplicates
        if _samefile(potential[0], match[0]):
            if verbose:
                sys.stderr.write("duplicate: %s (%s)\n" % potential)
            return None
    else:
        if not stat.S_ISREG(os.stat(potential[0]).st_mode):
            if verbose:
                sys.stderr.write("not a regular file: %s (%s)\n" % potential)
        elif not os.access(potential[0], os.X_OK):
            if verbose:
                sys.stderr.write("no executable access: %s (%s)\n"
                                 % potential)
        else:
            matches.append(potential)
            return potential

        
# ---- module API

def whichgen(command, path=None, verbose=0, exts=None):
    """Return a generator of full paths to the given command.
    
    "command" is a the name of the executable to search for.
    "path" is an optional alternate path list to search. The default it
        to use the PATH environment variable.
    "verbose", if true, will cause a 2-tuple to be returned for each
        match. The second element is a textual description of where the
        match was found.
    "exts" optionally allows one to specify a list of extensions to use
        instead of the standard list for this system. This can
        effectively be used as an optimization to, for example, avoid
        stat's of "foo.vbs" when searching for "foo" and you know it is
        not a VisualBasic script but ".vbs" is on PATHEXT. This option
        is only supported on Windows.

    This method returns a generator which yields either full paths to
    the given command or, if verbose, tuples of the form (<path to
    command>, <where path found>).
    """
    matches = []
    if path is None:
        usingGivenPath = 0
        path = os.environ.get("PATH", "").split(os.pathsep)
        if sys.platform.startswith("win"):
            path.insert(0, os.curdir)  # implied by Windows shell
    else:
        usingGivenPath = 1

    # Windows has the concept of a list of extensions (PATHEXT env var).
    if sys.platform.startswith("win"):
        if exts is None:
            exts = os.environ.get("PATHEXT", "").split(os.pathsep)
            # If '.exe' is not in exts then obviously this is Win9x and
            # or a bogus PATHEXT, then use a reasonable default.
            for ext in exts:
                if ext.lower() == ".exe":
                    break
            else:
                exts = ['.COM', '.EXE', '.BAT']
        elif not isinstance(exts, list):
            raise TypeError("'exts' argument must be a list or None")
    else:
        if exts is not None:
            raise WhichError("'exts' argument is not supported on "
                             "platform '%s'" % sys.platform)
        exts = []

    # File name cannot have path separators because PATH lookup does not
    # work that way.
    if os.sep in command or os.altsep and os.altsep in command:
        pass
    else:
        for i in range(len(path)):
            dirName = path[i]
            # On windows the dirName *could* be quoted, drop the quotes
            if sys.platform.startswith("win") and len(dirName) >= 2\
               and dirName[0] == '"' and dirName[-1] == '"':
                dirName = dirName[1:-1]
            for ext in ['']+exts:
                absName = os.path.abspath(
                    os.path.normpath(os.path.join(dirName, command+ext)))
                if os.path.isfile(absName):
                    if usingGivenPath:
                        fromWhere = "from given path element %d" % i
                    elif not sys.platform.startswith("win"):
                        fromWhere = "from PATH element %d" % i
                    elif i == 0:
                        fromWhere = "from current directory"
                    else:
                        fromWhere = "from PATH element %d" % (i-1)
                    match = _cull((absName, fromWhere), matches, verbose)
                    if match:
                        if verbose:
                            yield match
                        else:
                            yield match[0]
        match = _getRegisteredExecutable(command)
        if match is not None:
            match = _cull(match, matches, verbose)
            if match:
                if verbose:
                    yield match
                else:
                    yield match[0]


def which(command, path=None, verbose=0, exts=None):
    """Return the full path to the first match of the given command on
    the path.
    
    "command" is a the name of the executable to search for.
    "path" is an optional alternate path list to search. The default is
        to use the PATH environment variable.
    "verbose", if true, will cause a 2-tuple to be returned. The second
        element is a textual description of where the match was found.
    "exts" optionally allows one to specify a list of extensions to use
        instead of the standard list for this system. This can
        effectively be used as an optimization to, for example, avoid
        stat's of "foo.vbs" when searching for "foo" and you know it is
        not a VisualBasic script but ".vbs" is on PATHEXT. This option
        is only supported on Windows.

    If no match is found for the command, an empty String is returned.
    """
    try:
        match = next(whichgen(command, path, verbose, exts))
    except StopIteration:
        return ''
    return match


def commandExists(command):
    path = which(command)
    return path != ''


def whichall(command, path=None, verbose=0, exts=None):
    """Return a list of full paths to all matches of the given command
    on the path.  

    "command" is a the name of the executable to search for.
    "path" is an optional alternate path list to search. The default it
        to use the PATH environment variable.
    "verbose", if true, will cause a 2-tuple to be returned for each
        match. The second element is a textual description of where the
        match was found.
    "exts" optionally allows one to specify a list of extensions to use
        instead of the standard list for this system. This can
        effectively be used as an optimization to, for example, avoid
        stat's of "foo.vbs" when searching for "foo" and you know it is
        not a VisualBasic script but ".vbs" is on PATHEXT. This option
        is only supported on Windows.
    """
    return list(whichgen(command, path, verbose, exts))



""" Module to host templates classes"""
import collections
import glob
import os
import tempfile
from datetime import datetime
from pyworkflow import SCIPION_JSON_TEMPLATES, Config
from pyworkflow.utils import greenStr


class Template:
    def __init__(self, pluginName, tempPath):
        self.pluginName = pluginName
        self.templateName = os.path.basename(tempPath).replace(SCIPION_JSON_TEMPLATES, "")
        self.templatePath = os.path.abspath(tempPath)
        self.description, self.content = self._parseTemplate()
        self.params = None
        self.projectName = None

    def __str__(self):
        return self.templatePath

    def getObjId(self):
        return self.pluginName + '-' + self.templateName

    def genProjectName(self):
        self.projectName = self.getObjId() + '-' + datetime.now().strftime("%y%m%d-%H%M%S")

    def replaceEnvVariables(self):
        self.content = (self.content % os.environ).split('~')

    def _parseTemplate(self):
        with open(self.templatePath, 'r') as myFile:
            allContents = myFile.read().splitlines()
            description, index = Template.getDescription(allContents)

        if not description:
            description = 'Not provided'
        content = ''.join(allContents[index:])
        return description, content

    @staticmethod
    def getDescription(strList):
        # Example of json.template file with description:
        # -----------------------------------------------
        # Here goes the description
        # Description
        # Another description line...
        # [
        #    {
        #       "object.className": "ProtImportMovies",
        #       "object.id": "2",...
        # -----------------------------------------------

        contents_start_1 = '['
        contents_start_2 = '{'
        description = []
        counter = 0
        nLines = len(strList)

        while counter + 1 < nLines:
            currentLine = strList[counter]
            nextLine = strList[counter + 1]
            if contents_start_1 not in currentLine:
                description.append(currentLine)
            else:
                if contents_start_2 in nextLine:
                    break
                else:
                    description.append(currentLine)
            counter += 1

        return ''.join(description), counter

    def parseContent(self):

        def paramStr2Param(fieldIndex, fieldString):
            fieldLst = fieldString.split('|')

            title = fieldLst[0]
            defaultValue = fieldLst[1] if len(fieldLst) >= 2 else None
            varType = fieldLst[2] if len(fieldLst) >= 3 else None
            alias = fieldLst[3] if len(fieldLst) >= 4 else None

            return TemplateParam(fieldIndex, title, defaultValue, varType, alias)

        # Fill each field in the template in order to prevent spreading in the form
        self.params = collections.OrderedDict()
        for index in range(1, len(self.content), 2):
            param = paramStr2Param(index, self.content[index])
            self.params[param.getTitle()] = param

    def createTemplateFile(self):

        # Where to write the json file.
        (fileHandle, path) = tempfile.mkstemp()

        self._replaceFields()

        finalJson = "".join(self.content)

        os.write(fileHandle, finalJson.encode())
        os.close(fileHandle)

        print("New workflow saved at " + path)

        return path

    def _replaceFields(self):

        for field in self.params.values():
            self.content[field.getIndex()] = field.getValue()

    def getParams(self):
        return self.params

    def setParamValue(self, alias, newValue):
        paramsSetted = 0
        for field in self.params.values():
            if field.getAlias() == alias:
                oldValue = field.getValue()
                field.setValue(newValue)
                if field.validate():
                    paramsSetted += 1
                    print(greenStr("%s set to %s") %
                          (field.getTitle(), str(newValue)))
                else:
                    field.setValue(oldValue)
                    raise Exception("%s is not compatible with %s(%s) parameter." % (newValue, field.getTitle(), alias))
        if not paramsSetted:
            raise Exception("Alias %s not recognized." %
                            (alias))
        return paramsSetted


class TemplateParam(object):
    def __init__(self, index, title, value=None, varType=None, alias=None):
        self._index = index
        self._title = title
        self._value = value
        self._type = varType
        self._alias = alias

    def getTitle(self):
        return self._title

    def getIndex(self):
        return self._index

    def getType(self):
        return self._type

    def getValue(self):
        return self._value

    def setValue(self, value):
        self._value = value

    def getAlias(self):
        return self._alias

    def validate(self):
        return Validations.check(self._value, self._type)


class Validations:

    """ FIELDS VALIDATION """
    """ FIELDS TYPES"""
    FIELD_TYPE_STR = "0"
    FIELD_TYPE_BOOLEAN = "1"
    FIELD_TYPE_PATH = "2"
    FIELD_TYPE_INTEGER = "3"
    FIELD_TYPE_DECIMAL = "4"

    @classmethod
    def check(cls, value, fieldType):
        if fieldType == cls.FIELD_TYPE_BOOLEAN:
            return cls.validBoolean(value)
        elif fieldType == cls.FIELD_TYPE_DECIMAL:
            return cls.validDecimal(value)
        elif fieldType == cls.FIELD_TYPE_INTEGER:
            return cls.validInteger(value)
        elif fieldType == cls.FIELD_TYPE_PATH:
            return cls.validPath(value)
        elif fieldType == cls.FIELD_TYPE_STR:
            return cls.validString(value)

        else:
            print("Type %s for %s not recognized. Review the template."
                  % (type, value))
            return

    @staticmethod
    def validString(value):
        return value is not None

    @staticmethod
    def validInteger(value):
        return value.isdigit()

    @staticmethod
    def validPath(value):
        return os.path.exists(value)

    @staticmethod
    def validDecimal(value):

        try:
            float(value)
            return True
        except Exception as e:
            return False

    @staticmethod
    def validBoolean(value):
        return value is True or value is False


class TemplateList:
    def __init__(self, templates=None):
        self.templates = templates if templates else []

    def addTemplate(self, t):
        self.templates.append(t)

    def genFromStrList(self, templateList):
        for t in templateList:
            parsedPath = t.split(os.path.sep)
            pluginName = parsedPath[parsedPath.index('templates') - 1]
            self.addTemplate(Template(pluginName, t))

    def sortListByPluginName(self):
        # Create a identifier with both plugin and template names to sort by both
        self.templates = sorted(self.templates, key=lambda template: '.' + template.getObjId()
                                if template.getObjId().startswith('local') else template.getObjId())

        return self

    def addScipionTemplates(self, tempId=None):
        # Check if there is any .json.template in the template folder
        # get the template folder (we only want it to be included once)
        templateFolder = Config.getExternalJsonTemplates()
        for templateName in glob.glob1(templateFolder,
                                       "*" + SCIPION_JSON_TEMPLATES):
            t = Template("local", os.path.join(templateFolder, templateName))
            if tempId is not None:
                if t.getObjId() == tempId:
                    self.addTemplate(t)
                    break
            else:
                self.addTemplate(t)

    def addPluginTemplates(self, tempId=None):
        """
        Get the templates provided by all plugins.
        :return: a list of templates
        """
        # Check if other plugins have json.templates
        domain = Config.getDomain()
        # Check if there is any .json.template in the template folder
        # get the template folder (we only want it to be included once)
        for pluginName, pluginModule in domain.getPlugins().items():
            tempListPlugin = pluginModule.Plugin.getTemplates()
            for t in tempListPlugin:
                if tempId is not None:
                    if t.getObjId() == tempId:
                        self.addTemplate(t)
                        break
                else:
                    self.addTemplate(t)




# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Definition of Mock objects to be used within the tests in the Mock Domain
"""

import os

import pyworkflow.object as pwobj

NO_INDEX = 0


class MockObject(pwobj.OrderedObject):
    """Base object for all Mock classes"""
    def __init__(self, **kwargs):
        pwobj.OrderedObject.__init__(self, **kwargs)

    def __str__(self):
        return self.getClassName()

    def getFiles(self):
        """ Get all filePaths """
        return None


class Complex(MockObject):
    """ Simple class used for tests here. """

    cGold = complex(1.0, 1.0)

    def __init__(self, imag=0., real=0., **args):
        MockObject.__init__(self, **args)
        self.imag = pwobj.Float(imag)
        self.real = pwobj.Float(real)
        # Create reference complex values

    def __str__(self):
        return '(%s, %s)' % (self.imag, self.real)

    def __eq__(self, other):
        return (self.imag == other.imag and
                self.real == other.real)

    def hasValue(self):
        return True

    @classmethod
    def createComplex(cls):
        """Create a Complex object and set
        values with cls.cGold standard"""
        c = Complex()  # Create Complex object and set values
        c.imag.set(cls.cGold.imag)
        c.real.set(cls.cGold.real)
        return c


class MockAcquisition(MockObject):
    """Acquisition information"""
    def __init__(self, **kwargs):
        MockObject.__init__(self, **kwargs)
        self._magnification = pwobj.Float(kwargs.get('magnification', None))
        # Microscope voltage in kV
        self._voltage = pwobj.Float(kwargs.get('voltage', None))
        # Spherical aberration in mm
        self._sphericalAberration = pwobj.Float(kwargs.get('sphericalAberration',
                                                           None))
        self._amplitudeContrast = pwobj.Float(kwargs.get('amplitudeContrast', None))
        self._doseInitial = pwobj.Float(kwargs.get('doseInitial', 0))
        self._dosePerFrame = pwobj.Float(kwargs.get('dosePerFrame', None))

    def copyInfo(self, other):
        self.copyAttributes(other, '_magnification', '_voltage',
                            '_sphericalAberration', '_amplitudeContrast',
                            '_doseInitial', '_dosePerFrame')

    def getMagnification(self):
        return self._magnification.get()

    def setMagnification(self, value):
        self._magnification.set(value)

    def getVoltage(self):
        return self._voltage.get()

    def setVoltage(self, value):
        self._voltage.set(value)

    def getSphericalAberration(self):
        return self._sphericalAberration.get()

    def setSphericalAberration(self, value):
        self._sphericalAberration.set(value)

    def getAmplitudeContrast(self):
        return self._amplitudeContrast.get()

    def setAmplitudeContrast(self, value):
        self._amplitudeContrast.set(value)

    def getDoseInitial(self):
        return self._doseInitial.get()

    def setDoseInitial(self, value):
        self._doseInitial.set(value)

    def getDosePerFrame(self):
        return self._dosePerFrame.get()

    def setDosePerFrame(self, value):
        self._dosePerFrame.set(value)

    def __str__(self):
        return "\n    mag=%s\n    volt= %s\n    Cs=%s\n    Q0=%s\n\n" % \
               (self._magnification.get(),
                self._voltage.get(),
                self._sphericalAberration.get(),
                self._amplitudeContrast.get())


class MockImageDim(pwobj.CsvList):
    """ Just a wrapper to a pwobj.CsvList to store image dimensions
    as X, Y and Z.
    """
    def __init__(self, x=None, y=None, z=None):
        pwobj.CsvList.__init__(self, pType=int)
        if x is not None and y is not None:
            self.append(x)
            self.append(y)
            if z is not None:
                self.append(z)

    def getX(self):
        return self[0]

    def getY(self):
        return self[1]

    def getZ(self):
        return self[2]

    def __str__(self):
        if self.isEmpty():
            s = 'No-Dim'
        else:
            s = '%d x %d' % (self.getX(), self.getY())
            if self.getZ() > 1:
                s += ' x %d' % self.getZ()
        return s


class MockImage(MockObject):
    """Represents an EM Image object"""
    def __init__(self, location=None, **kwargs):
        """
         Params:
        :param location: Could be a valid location: (index, filename)
        or  filename
        """
        MockObject.__init__(self, **kwargs)
        # Image location is composed by an index and a filename
        self._index = pwobj.Integer(0)
        self._filename = pwobj.String()
        self._samplingRate = pwobj.Float()
        self._ctfModel = None
        self._acquisition = None
        if location:
            self.setLocation(location)

    def getSamplingRate(self):
        """ Return image sampling rate. (A/pix) """
        return self._samplingRate.get()

    def setSamplingRate(self, sampling):
        self._samplingRate.set(sampling)

    def getFormat(self):
        pass

    def getDataType(self):
        pass

    def getDimensions(self):
        """getDimensions is redundant here but not in setOfVolumes
         create it makes easier to create protocols for both images
         and sets of images
        """
        return self.getDim()

    def getDim(self):
        return 0, 0, 0  # We don't need dim now

    def getXDim(self):
        return self.getDim()[0] if self.getDim() is not None else 0

    def getYDim(self):
        return self.getDim()[1] if self.getDim() is not None else 0

    def getIndex(self):
        return self._index.get()

    def setIndex(self, index):
        self._index.set(index)

    def getFileName(self):
        """ Use the _objValue attribute to store filename. """
        return self._filename.get()

    def setFileName(self, filename):
        """ Use the _objValue attribute to store filename. """
        self._filename.set(filename)

    def getLocation(self):
        """ This function return the image index and filename.
        It will only differs from getFileName, when the image
        is contained in a stack and the index make sense.
        """
        return self.getIndex(), self.getFileName()

    def setLocation(self, *args):
        """ Set the image location, see getLocation.
        Params:
            First argument can be:
             1. a tuple with (index, filename)
             2. a index, this implies a second argument with filename
             3. a filename, this implies index=NO_INDEX
        """
        first = args[0]
        t = type(first)
        if t == tuple:
            index, filename = first
        elif t == int:
            index, filename = first, args[1]
        elif t == str:
            index, filename = NO_INDEX, first
        else:
            raise Exception('setLocation: unsupported type %s as input.' % t)

        self.setIndex(index)
        self.setFileName(filename)

    def getBaseName(self):
        return os.path.basename(self.getFileName())

    def copyInfo(self, other):
        """ Copy basic information """
        self.copyAttributes(other, '_samplingRate')

    def copyLocation(self, other):
        """ Copy location index and filename from other image. """
        self.setIndex(other.getIndex())
        self.setFileName(other.getFileName())

    def hasCTF(self):
        return self._ctfModel is not None

    def getCTF(self):
        """ Return the CTF model """
        return self._ctfModel

    def setCTF(self, newCTF):
        self._ctfModel = newCTF

    def hasAcquisition(self):
        return (self._acquisition is not None and
                self._acquisition.getVoltage() is not None and
                self._acquisition.getMagnification() is not None
                )

    def getAcquisition(self):
        return self._acquisition

    def setAcquisition(self, acquisition):
        self._acquisition = acquisition

    def hasTransform(self):
        return self._transform is not None

    def getTransform(self):
        return self._transform

    def setTransform(self, newTransform):
        self._transform = newTransform

    def hasOrigin(self):
        return self._origin is not None

    def getOrigin(self, force=False):
        """shifts in A"""
        if self.hasOrigin():
            return self._origin
        else:
            if force:
                return self._getDefaultOrigin()
            else:
                return None

    def _getDefaultOrigin(self):
        # original code from em required Transform, which will require
        # Matrix and then import numpy. Since this method is not used in
        # the tests I'm emptying it.
        pass

    def getShiftsiftsFromOrigin(self):
        origin = self.getOrigin(force=True).getShifts()
        x = origin[0]
        y = origin[1]
        z = origin[2]
        return x, y, z
        # x, y, z are floats in Angstroms

    def setShiftsInOrigin(self, x, y, z):
        origin = self.getOrigin(force=True)
        origin.setShifts(x, y, z)

    def setOrigin(self, newOrigin):
        """shifts in A"""
        self._origin = newOrigin

    def originResampled(self, originNotResampled, oldSampling):
        factor = self.getSamplingRate() / oldSampling
        shifts = originNotResampled.getShifts()
        origin = self.getOrigin(force=True)
        origin.setShifts(shifts[0] * factor,
                         shifts[1] * factor,
                         shifts[2] * factor)
        return origin

    def __str__(self):
        """ String representation of an Image. """
        dim = self.getDim()
        dimStr = str(MockImageDim(*dim)) if dim else 'No-Dim'
        return ("%s (%s, %0.2f Å/px)"
                % (self.getClassName(), dimStr,
                   self.getSamplingRate() or 99999.))

    def getFiles(self):
        filePaths = set()
        filePaths.add(self.getFileName())
        return filePaths


class MockMicrograph(MockImage):
    """ Represents an EM Micrograph object """
    def __init__(self, location=None, **kwargs):
        MockImage.__init__(self, location, **kwargs)
        self._micName = pwobj.String()

    def setMicName(self, micName):
        self._micName.set(micName)

    def getMicName(self):
        if self._micName.get():
            return self._micName.get()
        else:
            self.getFileName()

    def copyInfo(self, other):
        """ Copy basic information """
        MockImage.copyInfo(self, other)
        self.setMicName(other.getMicName())


class MockParticle(MockImage):
    """ Represents an EM Particle object """
    def __init__(self, location=None, **kwargs):
        MockImage.__init__(self, location, **kwargs)
        # This may be redundant, but make the Particle
        # object more indenpent for tracking coordinates
        self._coordinate = None
        self._micId = pwobj.Integer()
        self._classId = pwobj.Integer()

    def hasCoordinate(self):
        return self._coordinate is not None

    def setCoordinate(self, coordinate):
        self._coordinate = coordinate

    def getCoordinate(self):
        return self._coordinate

    def scaleCoordinate(self, factor):
        self.getCoordinate().scale(factor)

    def getMicId(self):
        """ Return the micrograph id if the coordinate is not None.
        or have set the _micId property.
        """
        if self._micId.hasValue():
            return self._micId.get()
        if self.hasCoordinate():
            return self.getCoordinate().getMicId()

        return None

    def setMicId(self, micId):
        self._micId.set(micId)

    def hasMicId(self):
        return self.getMicId() is not None

    def getClassId(self):
        return self._classId.get()

    def setClassId(self, classId):
        self._classId.set(classId)

    def hasClassId(self):
        return self._classId.hasValue()


class MockSet(pwobj.Set, MockObject):

    def _loadClassesDict(self):
        from pyworkflow.plugin import Domain
        classDict = Domain.getObjects()
        classDict.update(globals())

        return classDict

    def copyInfo(self, other):
        """ Define a dummy copyInfo function to be used
        for some generic operations on sets.
        """
        pass

    def clone(self):
        """ Override the clone defined in Object
        to avoid copying _mapperPath property
        """
        pass

    def copyItems(self, otherSet,
                  updateItemCallback=None,
                  itemDataIterator=None,
                  copyDisabled=False):
        """ Copy items from another set.
        If the updateItemCallback is passed, it will be
        called with each item (and optionally with a data row).
        This is a place where items can be updated while copying.
        This is useful to set new attributes or update values
        for each item.
        """
        for item in otherSet:
            # copy items if enabled or copyDisabled=True
            if copyDisabled or item.isEnabled():
                newItem = item.clone()
                if updateItemCallback:
                    row = None if itemDataIterator is None \
                        else next(itemDataIterator)
                    updateItemCallback(newItem, row)
                # If updateCallBack function returns attribute
                # _appendItem to False do not append the item
                if getattr(newItem, "_appendItem", True):
                    self.append(newItem)
            else:
                if itemDataIterator is not None:
                    next(itemDataIterator)  # just skip disabled data row

    def getFiles(self):
        return pwobj.Set.getFiles(self)


class MockSetOfImages(MockSet):
    """ Represents a set of Images """
    ITEM_TYPE = MockImage

    def __init__(self, **kwargs):
        MockSet.__init__(self, **kwargs)
        self._samplingRate = pwobj.Float()
        self._hasCtf = pwobj.Boolean(kwargs.get('ctf', False))
        self._isPhaseFlipped = pwobj.Boolean(False)
        self._isAmplitudeCorrected = pwobj.Boolean(False)
        self._acquisition = MockAcquisition()
        self._firstDim = MockImageDim()  # Dimensions of the first image

    def getAcquisition(self):
        return self._acquisition

    def setAcquisition(self, acquisition):
        self._acquisition = acquisition

    def hasAcquisition(self):
        return self._acquisition.getMagnification() is not None

    def append(self, image):
        """ Add a image to the set. """
        # If the sampling rate was set before, the same value
        # will be set for each image added to the set
        if self.getSamplingRate() or not image.getSamplingRate():
            image.setSamplingRate(self.getSamplingRate())
        # Copy the acquistion from the set to images
        # only override image acquisition if setofImages acquisition
        # is not none
        if self.hasAcquisition():
            # TODO: image acquisition should not be overwritten
            if not image.hasAcquisition():
                image.setAcquisition(self.getAcquisition())
        # Store the dimensions of the first image, just to
        # avoid reading image files for further queries to dimensions
        # only check this for first time append is called
        if self.isEmpty():
            self._setFirstDim(image)

        MockSet.append(self, image)

    def _setFirstDim(self, image):
        """ Store dimensions when the first image is found.
        This function should be called only once, to avoid reading
        dimension from image file. """
        if self._firstDim.isEmpty():
            self._firstDim.set(image.getDim())

    def copyInfo(self, other):
        """ Copy basic information (sampling rate and ctf)
        from other set of images to current one"""
        self.copyAttributes(other, '_samplingRate', '_isPhaseFlipped',
                            '_isAmplitudeCorrected', '_alignment')
        self._acquisition.copyInfo(other._acquisition)

    def getFiles(self):
        filePaths = set()
        uniqueFiles = self.aggregate(['count'], '_filename', ['_filename'])

        for row in uniqueFiles:
            filePaths.add(row['_filename'])
        return filePaths

    def setDownsample(self, downFactor):
        """ Update the values of samplingRate and scannedPixelSize
        after applying a downsampling factor of downFactor.
        """
        self.setSamplingRate(self.getSamplingRate() * downFactor)

    def setSamplingRate(self, samplingRate):
        """ Set the sampling rate and adjust the scannedPixelSize. """
        self._samplingRate.set(samplingRate)

    def getSamplingRate(self):
        return self._samplingRate.get()

    def getDim(self):
        """ Return the dimensions of the first image in the set. """
        if self._firstDim.isEmpty():
            return None
        x, y, z = self._firstDim
        return x, y, z

    def setDim(self, newDim):
        self._firstDim.set(newDim)

    def getXDim(self):
        return self.getDim()[0] if self.getDim() is not None else 0

    def isOddX(self):
        """ Return True if the first item x dimension is odd. """
        return self.getXDim() % 2 == 1

    def getDimensions(self):
        """Return first image dimensions as a tuple: (xdim, ydim, zdim)"""
        return self.getFirstItem().getDim()

    def __str__(self):
        """ String representation of a set of images. """
        sampling = self.getSamplingRate()

        if not sampling:
            print("FATAL ERROR: Object %s has no sampling rate!!!"
                  % self.getName())
            sampling = -999.0

        s = "%s (%d items, %s, %0.2f Å/px%s)" % \
            (self.getClassName(), self.getSize(),
             self._dimStr(), sampling, self._appendStreamState())
        return s

    def _dimStr(self):
        """ Return the string representing the dimensions. """
        return str(self._firstDim)

    def iterItems(self, orderBy='id', direction='ASC', where='1', limit=None):
        """ Redefine iteration to set the acquisition to images. """
        for img in pwobj.Set.iterItems(self, orderBy=orderBy, direction=direction,
                                       where=where, limit=limit):

            # Sometimes the images items in the set could
            # have the acquisition info per data row and we
            # don't want to override with the set acquisition for this case
            if not img.hasAcquisition():
                img.setAcquisition(self.getAcquisition())
            yield img

    def appendFromImages(self, imagesSet):
        """ Iterate over the images and append
        every image that is enabled.
        """
        for img in imagesSet:
            if img.isEnabled():
                self.append(img)

    def appendFromClasses(self, classesSet):
        """ Iterate over the classes and the element inside each
        class and append to the set all that are enabled.
        """
        for cls in classesSet:
            if cls.isEnabled() and cls.getSize() > 0:
                for img in cls:
                    if img.isEnabled():
                        self.append(img)


class MockSetOfMicrographs(MockSetOfImages):
    """ Create a base class for both Micrographs and Movies,
    but avoid to select Movies when Micrographs are required.
    """
    ITEM_TYPE = MockMicrograph

    def __init__(self, **kwargs):
        MockSetOfImages.__init__(self, **kwargs)
        self._scannedPixelSize = pwobj.Float()

    def copyInfo(self, other):
        """ Copy basic information (voltage, spherical aberration and
        sampling rate) from other set of micrographs to current one.
        """
        MockSetOfImages.copyInfo(self, other)
        self._scannedPixelSize.set(other.getScannedPixelSize())

    def setSamplingRate(self, samplingRate):
        """ Set the sampling rate and adjust the scannedPixelSize. """
        self._samplingRate.set(samplingRate)
        mag = self._acquisition.getMagnification()
        if mag is None:
            self._scannedPixelSize.set(None)
        else:
            self._scannedPixelSize.set(1e-4 * samplingRate * mag)

    def getScannedPixelSize(self):
        return self._scannedPixelSize.get()

    def setScannedPixelSize(self, scannedPixelSize):
        """ Set scannedPixelSize and update samplingRate. """
        mag = self._acquisition.getMagnification()
        if mag is None:
            raise Exception("SetOfMicrographs: cannot set scanned pixel size "
                            "if Magnification is not set.")
        self._scannedPixelSize.set(scannedPixelSize)
        self._samplingRate.set((1e+4 * scannedPixelSize) / mag)


class MockSetOfParticles(MockSetOfImages):
    """ Represents a set of Particles.
    The purpose of this class is to separate the
    concepts of Micrographs and Particles, even if
    both are considered Images
    """
    ITEM_TYPE = MockParticle
    REP_TYPE = MockParticle

    def __init__(self, **kwargs):
        MockSetOfImages.__init__(self, **kwargs)
        self._coordsPointer = pwobj.Pointer()

    def hasCoordinates(self):
        return self._coordsPointer.hasValue()

    def getCoordinates(self):
        """ Returns the SetOfCoordinates associated with
        this SetOfParticles"""
        return self._coordsPointer.get()

    def setCoordinates(self, coordinates):
        """ Set the SetOfCoordinates associates with
        this set of particles.
         """
        self._coordsPointer.set(coordinates)

    def copyInfo(self, other):
        """ Copy basic information (voltage, spherical aberration and
        sampling rate) from other set of micrographs to current one.
        """
        MockSetOfImages.copyInfo(self, other)
        self.setHasCTF(other.hasCTF())


class MockCoordinate(MockObject):
    """This class holds the (x,y) position and other information
    associated with a coordinate"""
    def __init__(self, **kwargs):
        MockObject.__init__(self, **kwargs)
        self._micrographPointer = pwobj.Pointer(objDoStore=False)
        self._x = pwobj.Integer(kwargs.get('x', None))
        self._y = pwobj.Integer(kwargs.get('y', None))
        self._micId = pwobj.Integer()
        self._micName = pwobj.String()

    def getX(self):
        return self._x.get()

    def setX(self, x):
        self._x.set(x)

    def shiftX(self, shiftX):
        self._x.sum(shiftX)

    def getY(self):
        return self._y.get()

    def setY(self, y):
        self._y.set(y)

    def shiftY(self, shiftY):
        self._y.sum(shiftY)

    def scale(self, factor):
        """ Scale both x and y coordinates by a given factor.
        """
        self._x.multiply(factor)
        self._y.multiply(factor)

    def getPosition(self):
        """ Return the position of the coordinate as a (x, y) tuple.
        mode: select if the position is the center of the box
        or in the top left corner.
        """
        return self.getX(), self.getY()

    def setPosition(self, x, y):
        self.setX(x)
        self.setY(y)

    def getMicrograph(self):
        """ Return the micrograph object to which
        this coordinate is associated.
        """
        return self._micrographPointer.get()

    def setMicrograph(self, micrograph):
        """ Set the micrograph to which this coordinate belongs. """
        self._micrographPointer.set(micrograph)
        self._micId.set(micrograph.getObjId())
        self._micName.set(micrograph.getMicName())

    def copyInfo(self, coord):
        """ Copy information from other coordinate. """
        self.setPosition(*coord.getPosition())
        self.setObjId(coord.getObjId())
        self.setBoxSize(coord.getBoxSize())

    def getMicId(self):
        return self._micId.get()

    def setMicId(self, micId):
        self._micId.set(micId)

    def invertY(self):
        if not self.getMicrograph() is None:
            dims = self.getMicrograph().getDim()
            height = dims[1]
            self.setY(height - self.getY())
        # else: error TODO

    def setMicName(self, micName):
        self._micName.set(micName)

    def getMicName(self):
        return self._micName.get()


class SetOfCoordinates(MockSet):
    """ Encapsulate the logic of a set of particles coordinates.
    Each coordinate has a (x,y) position and is related to a Micrograph
    The SetOfCoordinates can also have information about TiltPairs.
    """
    ITEM_TYPE = MockCoordinate

    def __init__(self, **kwargs):
        MockSet.__init__(self, **kwargs)
        self._micrographsPointer = pwobj.Pointer()
        self._boxSize = pwobj.Integer()

    def getBoxSize(self):
        """ Return the box size of the particles.
        """
        return self._boxSize.get()

    def setBoxSize(self, boxSize):
        """ Set the box size of the particles. """
        self._boxSize.set(boxSize)

    def iterMicrographs(self):
        """ Iterate over the micrographs set associated with this
        set of coordinates.
        """
        return self.getMicrographs()

    def iterMicrographCoordinates(self, micrograph):
        """ Iterates over the set of coordinates belonging to that micrograph.
        """
        pass

    def iterCoordinates(self, micrograph=None):
        """ Iterate over the coordinates associated with a micrograph.
        If micrograph=None, the iteration is performed over the whole
        set of coordinates.
        """
        if micrograph is None:
            micId = None
        elif isinstance(micrograph, int):
            micId = micrograph
        elif isinstance(micrograph, MockMicrograph):
            micId = micrograph.getObjId()
        else:
            raise Exception('Invalid input micrograph of type %s'
                            % type(micrograph))

        # Iterate over all coordinates if micId is None,
        # otherwise use micId to filter the where selection
        coordWhere = '1' if micId is None else '_micId=%d' % micId

        for coord in self.iterItems(where=coordWhere):
            yield coord

    def getMicrographs(self):
        """ Returns the SetOfMicrographs associated with
        this SetOfCoordinates"""
        return self._micrographsPointer.get()

    def setMicrographs(self, micrographs):
        """ Set the micrographs associated with this set of coordinates.
        Params:
            micrographs: Either a SetOfMicrographs object or a pointer to it.
        """
        if micrographs.isPointer():
            self._micrographsPointer.copy(micrographs)
        else:
            self._micrographsPointer.set(micrographs)
        
    def getFiles(self):
        filePaths = set()
        filePaths.add(self.getFileName())
        return filePaths

    def __str__(self):
        """ String representation of a set of coordinates. """
        if self._boxSize.hasValue():
            boxSize = self._boxSize.get()
            boxStr = ' %d x %d' % (boxSize, boxSize)
        else:
            boxStr = 'No-Box'
        s = "%s (%d items, %s%s)" % (self.getClassName(), self.getSize(),
                                     boxStr, self._appendStreamState())

        return s

    def copyInfo(self, other):
        """ Copy basic information (boxsize)
                from other set of coordinates to current one"""
        self.copyAttributes(other, '_boxSize')




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import datetime as dt
from collections import OrderedDict

import pyworkflow.object as pwobj
import pyworkflow.tests as pwtests
from ..objects import (Complex, MockSetOfImages, MockImage, MockObject,
                       MockAcquisition, MockMicrograph)


class ListContainer(pwobj.Object):
    def __init__(self, **args):
        pwobj.Object.__init__(self, **args)
        self.csv = pwobj.CsvList()
    
    
class TestObject(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestOutput(cls)

    def test_ObjectsDict(self):
        # Validate that the object dict is populated correctly
        basicObjNames = [
            'Object', 'Scalar', 'Integer', 'Float', 'String', 'Pointer',
            'OrderedObject', 'List', 'CsvList', 'PointerList', 'Set'
        ]
        self.assertTrue(all(name in pwobj.OBJECTS_DICT
                            for name in basicObjNames))

    def test_Object(self):
        value = 2
        i = pwobj.Integer(value)
        self.assertEqual(value, i.get())
        # compare objects
        i2 = pwobj.Integer(value)
        self.assertEqual(i, i2)
        
        value = 2.
        f = pwobj.Float(value)
        self.assertAlmostEqual(value, f.get())
        
        f.multiply(5)
        self.assertAlmostEqual(value*5, f.get())
        
        a = pwobj.Integer()
        self.assertEqual(a.hasValue(), False)
        c = Complex.createComplex()
        # Check values are correct
        self.assertEqual(c.imag.get(), Complex.cGold.imag)
        self.assertEqual(c.real.get(), Complex.cGold.real)
        
        # Test Boolean logic
        b = pwobj.Boolean(False)
        self.assertTrue(not b.get())

        b.set('True')
        self.assertTrue(b.get())
        
        b = pwobj.Boolean()
        b.set(False)
        self.assertTrue(not b.get())
        
        # CsvList should be empty if set to ''
        l = pwobj.CsvList()
        l.set('')
        self.assertEqual(len(l), 0)

        # Test emptiness
        self.assertIsNotEmpty(b)

    def testWithPointer(self):

        obj = pwobj.Integer(10)

        self.assertFalse(obj.hasPointer(),
                         "Default instantiation off Integer has a pointer.")

        self.assertEqual(obj.get(), 10, "Integer.get(), without pointer fails.")

        pointee = pwobj.Object()
        setattr(pointee, "value", pwobj.Integer(20))

        # Set a pointer (not a real case though, but enough here)
        obj.setPointer(pwobj.Pointer(pointee, extended='value'))

        self.assertEqual(obj.get(), 20, "Integer.get() fails with a pointer.")

    def test_String(self):
        value = 'thisisanstring'
        s = pwobj.String(value)
        self.assertEqual(value, s.get())
        self.assertEqual(s.hasValue(), True)
        
        s2 = pwobj.String()
        # None value is considered empty
        self.assertTrue(s2.empty(), "s2 string should be empty if None")
        s2.set(' ')
        # Only spaces is also empty
        self.assertTrue(s2.empty(), "s2 string should be empty if only spaces")
        s2.set('something')
        # No empty after some value
        self.assertFalse(s2.empty(),
                         "s2 string should not be empty after value")

        now = dt.datetime.now()
        s.set(now)
        self.assertEqual(now, s.datetime())

    def test_Pointer(self):
        c = Complex.createComplex()
        p = pwobj.Pointer()
        p.set(c)
        p.setExtended('Name')
        c.Name = pwobj.String('Paquito')

        self.assertEqual(p.get(), 'Paquito')
        stackFn = "images.stk"
        mrcsFn = "images.mrcs"
        fn = self.getOutputPath('test_images.sqlite')
        imgSet = MockSetOfImages(filename=fn)
        imgSet.setSamplingRate(1.0)
        for i in range(10):
            img = MockImage()
            img.setLocation(i+1, stackFn)
            imgSet.append(img)

        imgSet.write()

        # Test that image number 7 is correctly retrieved
        # from the set
        img7 = imgSet[7]
        self.assertEqual(img7.getFileName(), stackFn)

        # Modify some properties of image 7 to test update
        img7.setFileName(mrcsFn)
        img7.setSamplingRate(2.0)
        imgSet.update(img7)
        # Write changes after the image 7 update
        imgSet.write()

        # Read again the set to be able to retrieve elements
        imgSet = MockSetOfImages(filename=fn)

        # Validate that image7 was properly updated
        img7 = imgSet[7]
        self.assertEqual(img7.getFileName(), mrcsFn)

        o = MockObject()

        o.pointer = pwobj.Pointer()
        o.pointer.set(imgSet)

        o.refC = o.pointer.get()
        attrNames = [k for k, a in o.getAttributes()]
        # Check that 'refC' should not appear in attributes
        # since it is only an "alias" to an existing pointed value
        self.assertNotIn('refC', attrNames)

        self.assertFalse(o.pointer.hasExtended(),
                         'o.pointer should not have extended at this point')

        o.pointer.setExtended(7)

        self.assertTrue(o.pointer.hasExtended())
        self.assertTrue(o.pointer.hasExtended())
        self.assertEqual(o.pointer.getExtended(), "7")

        # Check that the Item 7 of the set is properly
        # retrieved by the pointer after setting the extended to 7
        self.assertEqual(imgSet[7], o.pointer.get())

        # Test the keyword arguments of Pointer contructor
        # repeat above tests with new pointer
        ptr = pwobj.Pointer(value=imgSet, extended=7)
        self.assertTrue(ptr.hasExtended())
        self.assertTrue(ptr.hasExtended())
        self.assertEqual(ptr.getExtended(), "7")

        # Check that the Item 7 of the set is properly
        # retrieved by the pointer after setting the extended to 7
        self.assertEqual(imgSet[7], ptr.get())

        o2 = pwobj.OrderedObject()
        o2.outputImages = imgSet
        ptr2 = pwobj.Pointer()
        ptr2.set(o2)
        # Test nested extended attributes
        ptr2.setExtended('outputImages.7')
        self.assertEqual(imgSet[7], ptr2.get())

        # Same as ptr2, but setting extended in constructor
        ptr3 = pwobj.Pointer(value=o2, extended='outputImages.7')
        self.assertEqual(imgSet[7], ptr3.get())

        # Test copy between pointer objects
        ptr4 = pwobj.Pointer()
        ptr4.copy(ptr3)
        self.assertEqual(imgSet[7], ptr4.get())
        self.assertEqual(ptr4.getExtended(), 'outputImages.7')

    def test_Sets(self):
        stackFn = "images.stk"
        fn = self.getOutputPath('test_images2.sqlite')
        print("Writing to sqlite: %s" % fn)

        imgSet = MockSetOfImages(filename=fn)
        imgSet.setSamplingRate(1.0)

        for i in range(10):
            img = MockImage()
            img.setLocation(i + 1, stackFn)
            imgSet.append(img)

        imgSet.write()

        # Test size is 10
        self.assertSetSize(imgSet, 10)

        # PERFORMANCE functionality
        def checkSetIteration(limit, skipRows=None):

            print("Checking set iteration with limit:%s and skipRows: %s"
                  % (limit, skipRows))
            expectedId = 1 if skipRows is None else skipRows+1
            index = 0
            for item in imgSet.iterItems(limit=(limit, skipRows)):
                self.assertEqual(item.getIndex(), expectedId+index,
                                 "Wrong item in set when using limits.")
                index += 1

            self.assertEqual(index, limit,
                             "Number of iterations wrong with limits")

        # Check iteration with limit
        checkSetIteration(2)

        # Check iteration with limit and skip rows
        checkSetIteration(3, 2)

    def test_copyAttributes(self):
        """ Check that after copyAttributes, the values
        were properly copied.
        """
        c1 = Complex(imag=10., real=11.)
        c2 = Complex(imag=0., real=1.0001)
        
        # Float values are different, should not be equal
        self.assertFalse(c1.equalAttributes(c2))
        c2.copyAttributes(c1, 'imag', 'real')
        
        self.assertTrue(c1.equalAttributes(c2), 
                        'Complex c1 and c2 have not equal attributes'
                        '\nc1: %s\nc2: %s\n' % (c1, c2))

        c1.score = pwobj.Float(1.)

        # If we copyAttributes again, score dynamic attribute should
        # be set in c2
        c2.copyAttributes(c1, 'score')
        self.assertTrue(hasattr(c2, 'score'))
        
    def test_equalAttributes(self):
        """ Check that equal attributes function behaves well
        to compare floats with a given precision.
        """
        c1 = Complex(imag=0., real=1.)
        c2 = Complex(imag=0., real=1.0001)
        
        # Since Float precision is 0.001, now c1 and c2
        # should have equal attributes
        self.assertTrue(c1.equalAttributes(c2))
        # Now if we set a more restrictive precision
        # c1 and c2 are not longer equals
        pwobj.Float.setPrecision(0.0000001)
        self.assertFalse(c1.equalAttributes(c2))

    def test_formatString(self):
        """ Test that Scalar objects behave well
        when using string formatting such as: %f or %d
        """
        i = pwobj.Integer(10)
        f = pwobj.Float(3.345)
        
        s1 = "i = %d, f = %0.3f" % (i, f)
        
        self.assertEqual(s1, "i = 10, f = 3.345")

    def test_getObjDict(self):
        """ Test retrieving an object dictionary with its attribute values."""
        acq1 = MockAcquisition(magnification=50000,
                               voltage=200,
                               sphericalAberration=2.7,
                               dosePerFrame=1)
        m1 = MockMicrograph(
            'my_movie.mrc', objId=1, objLabel='test micrograph',
            objComment='Testing store and retrieve from dict.')
        m1.setSamplingRate(1.6)
        m1.setAcquisition(acq1)
        m1Dict = m1.getObjDict(includeBasic=True)
        for k, v in m1Dict.items():
            if isinstance(v, str):
                v = "'%s'" % v
            print("('%s', %s)," % (k, v))

        goldDict1 = OrderedDict([
            ('object.id', 1),
            ('object.label', 'test micrograph'),
            ('object.comment', 'Testing store and retrieve from dict.'),
            ('_index', 0),
            ('_filename', 'my_movie.mrc'),
            ('_samplingRate', 1.6),
            ('_micName', None),
            ('_acquisition', None),
            ('_acquisition._magnification', 50000.0),
            ('_acquisition._voltage', 200.0),
            ('_acquisition._sphericalAberration', 2.7),
            ('_acquisition._amplitudeContrast', None),
            ('_acquisition._doseInitial', 0.0),
            ('_acquisition._dosePerFrame', 1.0),
             ])

        self.assertEqual(goldDict1, m1Dict)

    def test_Dict(self):
        d = pwobj.Dict(default='missing')
        d.update({1: 'one', 2: 'two'})

        # Return default value for any non-present key
        self.assertEqual('missing', d[10])

        # Return true for any 'contains' query
        self.assertTrue(100 in d)
        

class TestUtils(pwtests.BaseTest):

    @classmethod
    def setUpClass(cls):
        pwtests.setupTestOutput(cls)
            
    def test_ListsFunctions(self):
        """ Test of some methods that retrieve lists from string. """
        from pyworkflow.utils import getListFromValues, getFloatListFromValues,\
            getBoolListFromValues
        
        results = [('2x1 2x2 4 5', None, getListFromValues,
                    ['1', '1', '2', '2', '4', '5']),
                   ('2x1 2x2 4 5', None, getFloatListFromValues,
                    [1., 1., 2., 2., 4., 5.]),
                   ('1 2 3x3 0.5', 8, getFloatListFromValues,
                    [1., 2., 3., 3., 3., 0.5, 0.5, 0.5]),
                   ('3x1 3x0 1', 8, getBoolListFromValues,
                    [True, True, True, False, False, False, True, True]),
                   ]
        for s, n, func, goldList in results:
            l = func(s, length=n)
            self.assertAlmostEqual(l, goldList)
            if n:
                self.assertEqual(n, len(l))
                
    def test_Environ(self):
        """ Test the Environ class with its utilities. """
        from pyworkflow.utils import Environ
        env = Environ({'PATH': '/usr/bin:/usr/local/bin',
                       'LD_LIBRARY_PATH': '/usr/lib:/usr/lib64'
                       })
        env1 = Environ(env)
        env1.set('PATH', '/usr/local/xmipp')
        self.assertEqual(env1['PATH'], '/usr/local/xmipp')
        self.assertEqual(env1['LD_LIBRARY_PATH'], env['LD_LIBRARY_PATH'])
        
        env2 = Environ(env)
        env2.set('PATH', '/usr/local/xmipp', position=Environ.BEGIN)
        self.assertEqual(env2['PATH'], '/usr/local/xmipp' + os.pathsep +
                         env['PATH'])
        self.assertEqual(env2['LD_LIBRARY_PATH'], env['LD_LIBRARY_PATH'])
        
        env3 = Environ(env)
        env3.update({'PATH': '/usr/local/xmipp', 
                     'LD_LIBRARY_PATH': '/usr/local/xmipp/lib'},
                    position=Environ.END)
        self.assertEqual(env3['PATH'], env['PATH'] + os.pathsep +
                         '/usr/local/xmipp')
        self.assertEqual(env3['LD_LIBRARY_PATH'], env['LD_LIBRARY_PATH'] +
                         os.pathsep + '/usr/local/xmipp/lib')



import ast
import importlib
import inspect
import os
import sys
import types

from .constants import *

HOME = os.path.abspath(os.path.dirname(__file__))
PYTHON = os.environ.get(SCIPION_PYTHON, SCIPION_PYTHON_DEFAULT)


def join(*paths):
    """ join paths from HOME . """
    return os.path.join(HOME, *paths)


__resourcesPath = [join('resources')]


def findResource(filename):
    from .utils.path import findFile
    return findFile(filename, *__resourcesPath)


def genNotesHeading():
    return SCIPION_NOTES_HEADING_MSG


def getAppsPath(*paths):
    return join(APPS, *paths)


def getSyncDataScript():
    return getAppsPath(PW_SYNC_DATA)


def getScheduleScript():
    return getAppsPath(PW_SCHEDULE_RUN)


def getPwProtMpiRunScript():
    return getAppsPath(PW_PROTOCOL_MPIRUN)


def getTestsScript():
    return getAppsPath(PW_RUN_TESTS)


def getViewerScript():
    return getAppsPath(PW_VIEWER)


def getPyworkflowPath():
    """ Returns the path where pyworkflow is"""
    return os.path.dirname(__file__)


def getModuleFolder(moduleName):
    """ Returns the path of a module without importing it"""
    spec = importlib.util.find_spec(moduleName)
    return os.path.dirname(spec.origin)


class Config:
    """ Main Config for pyworkflow. It contains the main configuration values
    providing default values or, if present, taking them from the environment.
    It has SCIPION_HOME, SCIPION_USER_DATA ...
    Necessary value is SCIPION_HOME and has to be present in the environment"""

    @staticmethod
    def __get(key, default):
        value = os.environ.get(key, default)
        # Expand user and variables if string value
        if isinstance(value, str):
            value = os.path.expandvars(os.path.expanduser(value))

        return value

    class Root:
        """ Simple helper to return path from a root. """
        def __init__(self, root):
            self._root = root

        def join(self, *path):
            # We need to consider variable in the config with ~
            expanded = os.path.expanduser(os.path.join(*path))

            # join will not join if expanded is absolute
            return os.path.join(self._root, expanded)

    # Home for scipion
    _get = __get.__func__
    SCIPION_HOME = os.path.abspath(_get(SCIPION_HOME_VAR, ''))
    _root = Root(SCIPION_HOME)
    _join = _root.join

    # SCIPION PATHS
    # Where to install software
    SCIPION_SOFTWARE = _join(_get('SCIPION_SOFTWARE', 'software'))

    # Where is the input data for tests...also where it will be downloaded
    SCIPION_TESTS = _join(_get('SCIPION_TESTS', os.path.join('data', 'tests')))

    # User dependent paths
    # Location for scipion projects
    SCIPION_USER_DATA = _get('SCIPION_USER_DATA', '~/ScipionUserData')

    # General purpose scipion tmp folder
    SCIPION_TMP = _get('SCIPION_TMP', _join(SCIPION_USER_DATA, 'tmp'))
    # LOGS PATHS
    # Path for Scipion logs
    SCIPION_LOGS = _get('SCIPION_LOGS', _join(SCIPION_USER_DATA,'logs'))

    # Get general log file path
    SCIPION_LOG = _join(SCIPION_LOGS, 'scipion.log')

    # Where the output of the tests will be stored
    SCIPION_TESTS_OUTPUT = _get('SCIPION_TESTS_OUTPUT', _join(SCIPION_USER_DATA, 'Tests'))

    SCIPION_SUPPORT_EMAIL = _get('SCIPION_SUPPORT_EMAIL', 'scipion@cnb.csic.es')

    SCIPION_LOGO = _get('SCIPION_LOGO', 'scipion_logo.gif')

    # Config folders
    SCIPION_CONFIG = _get('SCIPION_CONFIG', 'scipion.conf')
    SCIPION_LOCAL_CONFIG = _get('SCIPION_LOCAL_CONFIG', SCIPION_CONFIG)
    SCIPION_HOSTS = _get('SCIPION_HOSTS', 'hosts.conf')
    SCIPION_PROTOCOLS = _get('SCIPION_PROTOCOLS', 'protocols.conf')

    SCIPION_PLUGIN_JSON = _get('SCIPION_PLUGIN_JSON', None)
    SCIPION_PLUGIN_REPO_URL = _get('SCIPION_PLUGIN_REPO_URL',
                                   'http://scipion.i2pc.es/getplugins/')

    # REMOTE Section
    SCIPION_URL = _get('SCIPION_URL' , 'http://scipion.cnb.csic.es/downloads/scipion')
    SCIPION_URL_SOFTWARE = _get('SCIPION_URL_SOFTWARE', SCIPION_URL + '/software')
    SCIPION_URL_TESTDATA = _get('SCIPION_URL_TESTDATA', SCIPION_URL + '/data/tests')

    # Scipion Notes
    SCIPION_NOTES_FILE = _get(SCIPION_NOTES_FILE, 'notes.txt')
    SCIPION_NOTES_PROGRAM = _get(SCIPION_NOTES_PROGRAM, None)
    SCIPION_NOTES_ARGS = _get(SCIPION_NOTES_ARGS, None)

    # Aspect
    SCIPION_FONT_NAME = _get('SCIPION_FONT_NAME', "Helvetica")
    SCIPION_FONT_SIZE = int(_get('SCIPION_FONT_SIZE', 10))

    # Notification
    SCIPION_NOTIFY = _get('SCIPION_NOTIFY', 'True')

    SCIPION_CWD = _get('SCIPION_CWD', os.path.abspath(os.getcwd()))

    try:
        VIEWERS = ast.literal_eval(_get('VIEWERS', "{}"))
    except Exception as e:
        VIEWERS = {}
        print("ERROR loading preferred viewers, VIEWERS variable will be ignored")
        print(e)

    SCIPION_DOMAIN = _get(SCIPION_DOMAIN, None)
    SCIPION_TESTS_CMD = _get(SCIPION_TESTS_CMD, getTestsScript())

    # ---- Getters ---- #
    # Getters are alternatives to offer a variable, but preventing it to be stored in the config
    @classmethod
    def getLibFolder(cls):
        """
        :return: Folder where libraries must be placed in case a binding needs them
        """
        lib = cls._join(cls.SCIPION_SOFTWARE, 'lib')
        os.makedirs(lib, exist_ok=True)
        return lib

    @classmethod
    def getBindingsFolder(cls):
        """
        Folder where bindings must be placed. This folder is added to sys.path at launching time.
        If the binding depends on a dynamic libraries, those must be placed at cls.getLibFolder()
        :return:   The bindings folder
        """
        bindings = cls._join(cls.SCIPION_SOFTWARE, 'bindings')
        os.makedirs(bindings, exist_ok=True)
        return bindings

    @classmethod
    def getVars(cls):
        """ Return a dictionary with all variables defined
        in this Config.
        """
        configVars = dict()
        # For each variable, also in base classes
        for baseCls in inspect.getmro(cls):
            for name, value in vars(baseCls).items():
                # Skip methods and internal attributes starting with __
                # (e.g __doc__, __module__, etc)
                if isinstance(value, str) and not name.startswith('__'):
                    configVars[name] = value
        return configVars

    @classmethod
    def printVars(cls):
        """ Print the variables dict, mostly for debugging. """
        from .utils import prettyDict
        prettyDict(cls.getVars())

    @classmethod
    def getDomain(cls):
        """ Import domain module from path or name defined in SCIPION_DOMAIN.
        """
        value = cls.SCIPION_DOMAIN

        if not value:
            return None

        if os.path.isdir(value):
            dirname, value = os.path.split(value)
            sys.path.append(dirname)

        return importlib.import_module(value).Domain

    @classmethod
    def setDomain(cls, moduleOrNameOrPath):
        if isinstance(moduleOrNameOrPath, types.ModuleType):
            value = os.path.abspath(moduleOrNameOrPath.__path__[0])
        else:
            value = moduleOrNameOrPath
        cls.SCIPION_DOMAIN = value
        os.environ[SCIPION_DOMAIN] = value

    @staticmethod
    def getPythonLibFolder():
        from sysconfig import get_paths
        return os.path.join(get_paths()['data'], "lib")

    @staticmethod
    def debugOn(*args):
        from .utils import envVarOn
        return bool(envVarOn(SCIPION_DEBUG, *args))

    @staticmethod
    def toggleDebug():
        newValue = not Config.debugOn()
        os.environ[SCIPION_DEBUG] = str(newValue)
        os.environ[SCIPION_DEBUG_NOCLEAN] = str(newValue)

    @classmethod
    def getExternalJsonTemplates(cls):
        return os.path.dirname(cls.SCIPION_CONFIG)


# Add bindings folder to sys.path
sys.path.append(Config.getBindingsFolder())
